/*
 *  auth_paula: authenticate SAMBA against XAD/DCOM PAULA
 *  Copyright (C) 2001-2004 PADL Software Pty Ltd. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 */
/*
 *  Unix SMB/CIFS implementation.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1998,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1998,
 *  Copyright (C) Paul Ashton                       1998.
 *  Copyright (C) Jeremy Allison                    1999.
 *  Copyright (C) Andrew Bartlett                   2003.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <includes.h>

#include "rpc_paula.h"

#undef DBGC_CLASS
#define DBGC_CLASS DBGC_AUTH

NTSTATUS init_module(void);

/*
 * Setup a PAULA NT logon request structure from the supplied parameters
 */
static void init_paula_q_nt_logon_request(PAULA_Q_NT_LOGON_REQUEST *q_p,
	uint32 cookie,
	char *domain,
	char *workstation,
	char *username,
	uchar challenge[8],
	uchar lm_response[24],
	uint32 nt_response_len,
	uchar *nt_response)
{
	q_p->cookie = cookie;

	q_p->domain.str_max_len      = strlen(domain) + 1;
	q_p->domain.offset           = 0;
	q_p->domain.str_str_len      = q_p->domain.str_max_len;
	q_p->domain.buffer           = domain;

	q_p->workstation.str_max_len = strlen(workstation) + 1;
	q_p->workstation.offset      = 0;
	q_p->workstation.str_str_len = q_p->workstation.str_max_len;
	q_p->workstation.buffer      = workstation;

	q_p->username.str_max_len    = strlen(username) + 1;
	q_p->username.offset         = 0;
	q_p->username.str_str_len    = q_p->username.str_max_len;
	q_p->username.buffer         = username;

	memcpy(q_p->challenge, challenge,   8);

	if (lm_response == NULL)
		memset(q_p->lm_response, 0, 24);
	else
		memcpy(q_p->lm_response, lm_response, 24);

	q_p->nt_response_len = nt_response_len;
	q_p->nt_response     = nt_response;
}

/*
 * Marshall a PAULA NT logon request RPC request PDU
 */
static BOOL paula_io_q_nt_logon_request(const char *desc, PAULA_Q_NT_LOGON_REQUEST *q_p,
	prs_struct *ps, int depth)
{
	if (!prs_uint32("cookie", ps, 0, &q_p->cookie))
		return False;

	if (!smb_io_string2("domain", &q_p->domain, 1, ps, 0))
		return False;

	if (!smb_io_string2("workstation", &q_p->workstation, 1, ps, 0))
		return False;

	if (!smb_io_string2("username", &q_p->username, 1, ps, 0))
		return False;

	if (!prs_uint8s(False, "challenge", ps, 0, q_p->challenge, 8))
		return False;

	if (!prs_uint8s(False, "lm_response", ps, 0, q_p->lm_response, 24))
		return False;

	if (!prs_align(ps))
		return False;

	if (!prs_uint32("nt_response_len", ps, 0, &q_p->nt_response_len))
		return False;

	if (!prs_uint32("nt_response_len_ndr", ps, 0, &q_p->nt_response_len))
		return False;

	if (UNMARSHALLING(ps)) {
		q_p->nt_response = prs_alloc_mem(ps, q_p->nt_response_len);
		if (!q_p->nt_response)
			return False;
	}

	if (!prs_uint8s(False, "nt_response", ps, 0, q_p->nt_response, q_p->nt_response_len))
		return False;

	return True;
}

/*
 * Marshall a PAULA NT logon request RPC response PDU
 */
static BOOL paula_io_r_nt_logon_request(const char *desc, PAULA_R_NT_LOGON_REQUEST *r_p,
	prs_struct *ps, int depth)
{
	if (!prs_uint32("cookie", ps, 0, &r_p->cookie))
		return False;
	
	if (!prs_uint32("status", ps, 0, &r_p->status))
		return False;

	if (!prs_uint32("sub_status", ps, 0, &r_p->sub_status))
		return False;

	if (!prs_uint32("result", ps, 0, &r_p->result))
		return False;

	if (!prs_uint8s(False, "user_session_key", ps, 0, r_p->user_session_key, 16))
		return False;

	if (!prs_uint8s(False, "lm_session_key", ps, 0, r_p->lm_session_key, 8))
		return False;

	return True;
}

/*
 * Open a socket to the local PAULA service
 */
static BOOL paula_open_socket(int *sd)
{
	struct sockaddr_un sunaddr;
	struct stat st;

	ZERO_STRUCT(sunaddr);
	sunaddr.sun_family = AF_UNIX;

	snprintf(sunaddr.sun_path, sizeof(sunaddr.sun_path), "%s", PAULA_DOMAIN_SOCKET);

	become_root();
	if (lstat(sunaddr.sun_path, &st) == -1) {
		unbecome_root();
		DEBUG(0, ("paula_open_socket: could not access %s\n", sunaddr.sun_path));
		return False;
	}

	if (!S_ISSOCK(st.st_mode)) {
		unbecome_root();
		DEBUG(0, ("paula_open_socket: %s not socket\n", sunaddr.sun_path));
		return False;
	}

	*sd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (*sd == -1) {
		unbecome_root();
		DEBUG(0, ("paula_open_socket: could not open socket - %s\n", strerror(errno)));
		return False;
	}

	if (connect(*sd, (struct sockaddr *)&sunaddr, sizeof(sunaddr)) == -1) {
		unbecome_root();
		DEBUG(0, ("paula_open_socket: could not connect to socket - %s\n", strerror(errno)));
		return False;
	}

	unbecome_root();

	return True;
}

/*
 * Create an anonymous RPC bind rqeuest PDU
 */
static BOOL paula_create_rpc_bind_req(prs_struct *rpc_out,
	uint32 rpc_call_id,
	RPC_IFACE *abstract,
	RPC_IFACE *transfer)
{
	RPC_HDR hdr;
	RPC_HDR_RB hdr_rb;
#if 0
	RPC_HDR_AUTH hdr_auth;
#endif
	int auth_len = 0;
#if 0
	int auth_type, auth_level;
	size_t saved_header_offset = 0;
#endif
	prs_struct auth_info;

	prs_init(&auth_info, RPC_HDR_AUTH_LEN, prs_get_mem_context(rpc_out), MARSHALL);

	/* Create the request RPC_HDR */
	init_rpc_hdr(&hdr, RPC_BIND, 0x3, rpc_call_id,
		RPC_HEADER_LEN + RPC_HDR_RB_LEN + prs_offset(&auth_info), auth_len);

	if (!smb_io_rpc_hdr("hdr    ", &hdr, rpc_out, 0)) {
		DEBUG(0, ("paula_create_rpc_bind_req: failed to marshall RPC_HDR.\n"));
		prs_mem_free(&auth_info);
		return False;
	}

	/* Create the bind request RPC_HDR_RB */
	init_rpc_hdr_rb(&hdr_rb, MAX_PDU_FRAG_LEN, MAX_PDU_FRAG_LEN, 0x0,
		0x1, 0x0, 0x1, abstract, transfer);

	/* Marshall the bind request data */
	if (!smb_io_rpc_hdr_rb("", &hdr_rb, rpc_out, 0)) {
		DEBUG(0, ("paula_create_rpc_bind_req: failed to marshall RPC_HDR_RB.\n"));
		prs_mem_free(&auth_info);
		return False;
	}

	prs_mem_free(&auth_info);
	return True;
}

/*
 * RPC call ID 
 */
static uint32 paula_get_rpc_call_id(void)
{       
	static uint32 call_id = 0;
	return ++call_id;
}

/*
 * Verify RPC header
 */
static BOOL paula_rpc_check_hdr(prs_struct *rdata, RPC_HDR *rhdr,
			  BOOL *first, BOOL *last, uint32 *len)
{
	DEBUG(5,("paula_rpc_check_hdr: rdata->data_size = %u\n", (uint32)prs_data_size(rdata) ));

	/* Next call sets endian bit. */

	if(!smb_io_rpc_hdr("rpc_hdr   ", rhdr, rdata, 0)) {
		DEBUG(0,("paula_rpc_check_hdr: Failed to unmarshall RPC_HDR.\n"));
		return False;
	}

	if (prs_offset(rdata) != RPC_HEADER_LEN) {
		DEBUG(0,("paula_rpc_check_hdr: offset was %x, should be %x.\n", prs_offset(rdata), RPC_HEADER_LEN));
		return False;
	}
	
	(*first) = ((rhdr->flags & RPC_FLG_FIRST) != 0);
	(*last) = ((rhdr->flags & RPC_FLG_LAST ) != 0);
	(*len) = (uint32)rhdr->frag_len - prs_data_size(rdata);
	
	return (rhdr->pkt_type != RPC_FAULT);
}

/*
 * Verify RPC bind response
 */	
static BOOL paula_check_bind_response(RPC_HDR_BA *hdr_ba, const int pipe_idx, RPC_IFACE *transfer)
{       
	if (hdr_ba->addr.len == 0) {
		DEBUG(4,("Ignoring length check -- AS/U bug (server didn't fill in the pipe name correctly)"));
	}


	/* check the transfer syntax */
	if ((hdr_ba->transfer.version != transfer->version) ||
	     (memcmp(&hdr_ba->transfer.uuid, &transfer->uuid, sizeof(transfer->uuid)) !=0)) {
		DEBUG(2,("bind_rpc_pipe: transfer syntax differs\n"));
		return False;
	}

	/* lkclXXXX only accept one result: check the result(s) */
	if (hdr_ba->res.num_results != 0x1 || hdr_ba->res.result != 0) {
		DEBUG(2,("bind_rpc_pipe: bind denied results: %d reason: %x\n",
			  hdr_ba->res.num_results, hdr_ba->res.reason));
	}

	DEBUG(5,("bind_rpc_pipe: accepted!\n"));
	return True;
}

/*
 * Make a remote procedure call on a raw socket
 */
static BOOL paula_rpc(int sd,
	uint16 max_xmit_frag,
	uint16 max_recv_frag,
	prs_struct *data,
	prs_struct *rdata,
	uint8 expected_pkt_type)
{
	uint32 len;
#if 0
	char *rparam = NULL;
	uint32 rparam_len = 0;
#endif
	BOOL first = True;
	BOOL last = True;
	RPC_HDR rhdr;
	char *pdata = data ? prs_data_p(data) : NULL;
	uint32 data_len = data ? prs_offset(data) : 0;
	char *prdata = NULL;
	uint32 current_offset = 0;
#if 0
	uint32 fragment_start = 0;
	int auth_padding_len = 0;
	uint32 max_data = max_xmit_frag ? max_xmit_frag : 1024;
#endif
	size_t data_to_read;
	ssize_t n;

	/* Send the RPC request */
	n = write_socket(sd, pdata, data_len);
	if (n != data_len) {
		DEBUG(0, ("paula_rpc: only wrote %d bytes of %d to socket\n", n, data_len));
		return False;
	}

	/* Read the RPC header first */
	if (!prs_init(rdata, RPC_HEADER_LEN, prs_get_mem_context(data), UNMARSHALL)) {
		DEBUG(0, ("paula_rpc: failed to initialize parse struct for response header\n"));
		return False;
	}

	n = sys_read(sd, prs_data_p(rdata), RPC_HEADER_LEN);
	if (n != RPC_HEADER_LEN) {
		DEBUG(0, ("paula_rpc: failed to read RPC_HDR\n"));
		prs_mem_free(rdata);
		return False;
	}

	current_offset = n;

	/* Parse the RPC header so we can determine the fragment length */
	if (!paula_rpc_check_hdr(rdata, &rhdr, &first, &last, &len)) {
		prs_mem_free(rdata);
		return False;
	}

	if (rhdr.pkt_type == RPC_BINDACK) {
		if (!last && !first) {
			DEBUG(5, ("paula_rpc: bug in server (AS/U?), setting fragment first/last ON.\n"));
			first = True;
			last = True;
		}
	}

	if (rhdr.pkt_type == RPC_BINDNACK) {
		DEBUG(3, ("Bind NACK received\n"));
		prs_mem_free(rdata);
		return False;
	}

	data_to_read = rhdr.frag_len - RPC_HEADER_LEN;

	/* Now, read the rest of the data */
	if (!prs_force_grow(rdata, data_to_read)) {
		DEBUG(0, ("paula_rpc: failed to grow buffer to receive remaining data in fragment\n"));
		prs_mem_free(rdata);
		return False;
	}

	prdata = prs_data_p(rdata) + current_offset;

	do
	{
		n = sys_read(sd, prdata, data_to_read);

		DEBUG(5, ("paula_rpc: num_read = %d, read offset: %d, to read: %d\n",
			n, current_offset, data_to_read));

		data_to_read -= n;
		current_offset += n;
		prdata += n;
	} while (n > 0 && data_to_read > 0);

	if (rhdr.pkt_type == RPC_RESPONSE) {
		RPC_HDR_RESP rhdr_resp;

		if (!smb_io_rpc_hdr_resp("rpc_hdr_resp", &rhdr_resp, rdata, 0)) {
			DEBUG(5, ("paula_rpc: failed to unmarshall RPC_HDR_RESP.\n"));
			prs_mem_free(rdata);
			return False;
		}
	}

	if (rhdr.pkt_type != expected_pkt_type) {
		DEBUG(3, ("paula_rpc: got unexpected packet type - %d, not %d\n",
			rhdr.pkt_type, expected_pkt_type));
		prs_mem_free(rdata);
		return False;
	}

	if (first && last) {
		DEBUG(6, ("paula_rpc: fragment first and last both set - returning\n"));
		return True;
	}

	/*
	 * Taking a short cut, I know, but a PAULA resposne should always fit
	 * within a single fragment.
	 */
	DEBUG(0, ("paula_rpc: response does not fit into a single fragment; fragment assembly not supported.\n"));

	prs_mem_free(rdata);

	return False;
}

/*
 * Bind to the local PAULA server
 */
static BOOL paula_bind(TALLOC_CTX *mem_ctx, int sd, uint16 *max_xmit_frag, uint16 *max_recv_frag)
{
	RPC_IFACE abstract = SYNT_PAULA_V1;
	RPC_IFACE transfer = TRANS_SYNT_V2;
	prs_struct rpc_out;
	prs_struct rdata;
	uint32 rpc_call_id;
	char buffer[MAX_PDU_FRAG_LEN];

	prs_init(&rpc_out, 0, mem_ctx, MARSHALL);
	prs_give_memory(&rpc_out, buffer, sizeof(buffer), False);
	rpc_call_id = paula_get_rpc_call_id();

	/* Marshall the outgoing data */
	paula_create_rpc_bind_req(&rpc_out, rpc_call_id,
		&abstract, &transfer);

	/* Initialize the incoming data struct */
	prs_init(&rdata, 0, mem_ctx, UNMARSHALL);

	/* Send data on socket; receive response */
	if (paula_rpc(sd, 0, 0, &rpc_out, &rdata, RPC_BINDACK)) {
		RPC_HDR_BA hdr_ba;

		DEBUG(5, ("paula_bind: paula_rpc returned OK.\n"));

		if (!smb_io_rpc_hdr_ba("", &hdr_ba, &rdata, 0)) {
			DEBUG(0, ("paula_bind: Failed to umarshall RPC_HDR_BA.\n"));
			prs_mem_free(&rdata);
			return False;
		}

		if (!paula_check_bind_response(&hdr_ba, 0, &transfer)) {
			DEBUG(2, ("paula_bind: check_bind_repsonse failed.\n"));
			prs_mem_free(&rdata);
			return False;
		}

		*max_xmit_frag = hdr_ba.bba.max_tsize;
		*max_recv_frag = hdr_ba.bba.max_rsize;

		prs_mem_free(&rdata);
		return True;
	}

	return False;
}

/*
 * Create a general RPC request
 */
static uint32 paula_create_rpc_request(prs_struct *rpc_out,
	uint8 op_num,
	int data_len,
	int auth_len,
	uint8 flags,
	uint32 oldid,
	uint32 data_left)
{
	uint32 alloc_hint;
	RPC_HDR hdr;
	RPC_HDR_REQ hdr_req;
	uint32 callid = oldid ? oldid : paula_get_rpc_call_id();

	DEBUG(5, ("paula_create_rpc_request: op_num 0x%x data_len: 0x%x\n", op_num, data_len));

	/* create the RPC header RPC_HDR */
	init_rpc_hdr(&hdr, RPC_REQUEST, flags, callid, data_len, auth_len);

	if (auth_len != 0)
		alloc_hint = data_len - RPC_HEADER_LEN - RPC_HDR_AUTH_LEN - auth_len;
	else
		alloc_hint = data_len - RPC_HEADER_LEN;

	DEBUG(10, ("paula_create_rpc_request: data_len: %x auth_len: %x alloc_hint: %x\n",
		data_len, auth_len, alloc_hint));

	/* create the RPC request RPC_HDR_REQ */
	init_rpc_hdr_req(&hdr_req, alloc_hint, op_num);

	if (!smb_io_rpc_hdr("hdr    ", &hdr, rpc_out, 0))
		return 0;

	if (!smb_io_rpc_hdr_req("hdr_req", &hdr_req, rpc_out, 0))
		return 0;

	if (prs_offset(rpc_out) != RPC_HEADER_LEN + RPC_HDR_REQ_LEN)
		return 0;

	return callid;
}

static BOOL paula_request(TALLOC_CTX *mem_ctx,
	int sd,
	uint16 max_xmit_frag,
	uint16 max_recv_frag,
	uint8 op_num,
	prs_struct *data,
	prs_struct *rdata)
{
	uint32 auth_len, real_auth_len, auth_hdr_len, max_data, data_left, data_sent;
	uint32 callid = 0;

	auth_len = 0;
	real_auth_len = 0;
	auth_hdr_len = 0;

	max_data = max_xmit_frag - RPC_HEADER_LEN - RPC_HDR_REQ_LEN
		- auth_hdr_len - auth_len - 8;

	for (data_left = prs_offset(data), data_sent = 0; data_left > 0;) {
		prs_struct outgoing_packet;
		uint32 data_len, send_size;
		uint8 flags = 0;
#if 0
		uint32 auth_padding = 0;
		DATA_BLOB sign_blob;
#endif

		send_size = MIN(data_left, max_data);

		data_len = RPC_HEADER_LEN + RPC_HDR_REQ_LEN + prs_offset(data);

		if (!prs_init(&outgoing_packet, data_len + 8, mem_ctx, MARSHALL)) {
			DEBUG(0, ("paula_request: Failed to malloc %u bytes.\n", data_len + 8));
			return False;
		}

		if (data_left == prs_offset(data))
			flags |= RPC_FLG_FIRST;

		if (data_left <= max_data)
			flags |= RPC_FLG_LAST;

		if (!(callid = paula_create_rpc_request(&outgoing_packet, op_num,
							data_len, real_auth_len, flags,
							callid, data_left))) {
			DEBUG(0, ("paula_request: Failed to create RPC request.\n"));
			prs_mem_free(&outgoing_packet);
			return False;
		}

		prs_append_prs_data(&outgoing_packet, data);
		DEBUG(100, ("data_len: %x data_calc_len: %x\n", data_len,
			prs_offset(&outgoing_packet)));

		if (!paula_rpc(sd, max_xmit_frag, max_recv_frag,
			&outgoing_packet, rdata,
			RPC_RESPONSE)) {
			prs_mem_free(&outgoing_packet);
			return False;
		}

		prs_mem_free(&outgoing_packet);
		data_sent += send_size;
		data_left -= send_size;
	}

#if 0
	prs_dump("reply_paula", op_num, rdata);
#endif

	/* Now we have the packet, get the response out of it and stick it in rdata */

	return True;
}

/*
 * Bind to the local PAULA server and make a RPC with the supplied data
 */
static BOOL paula_bind_and_request(TALLOC_CTX *mem_ctx,
	uint8 op_num,
	prs_struct *qdata,
	prs_struct *rdata)
{
	BOOL ret = False;
	int sd = -1;
	uint16 max_xmit_frag = 0;
	uint16 max_recv_frag = 0;

	if (!paula_open_socket(&sd))
		goto done;

	/* now format the bind request */
	if (!paula_bind(mem_ctx, sd, &max_xmit_frag, &max_recv_frag))
		goto done;

	/* now send the RPC */
	if (!paula_request(mem_ctx, sd, max_xmit_frag, max_recv_frag, op_num, qdata, rdata))
		goto done;

	ret = True;

done:
	if (sd > 0)
		close(sd);

	return ret;
}

/*
 * Make an NT logon request RPC to the local PAULA server
 */
static void paula_nt_logon_request(TALLOC_CTX *mem_ctx,
	uint32 cookie_in,
	char *domain,
	char *workstation,
	char *username,
	uchar challenge[8],
	uchar lm_response[24],
	uint32 nt_response_len,
	uchar *nt_response,
	uint32 *cookie_out,
	uint32 *status,
	uint32 *sub_status,
	uint32 *result,
	uchar user_session_key[16],
	uchar lm_session_key[8])
{
	prs_struct qbuf, rbuf;
	PAULA_Q_NT_LOGON_REQUEST q;
	PAULA_R_NT_LOGON_REQUEST r;
	BOOL ret = False;

	ZERO_STRUCT(q);
	ZERO_STRUCT(r);

	*status = NT_STATUS_V(NT_STATUS_UNSUCCESSFUL);
	*sub_status = NT_STATUS_V(NT_STATUS_UNSUCCESSFUL);
	*result = 0;

	prs_init(&qbuf, MAX_PDU_FRAG_LEN, mem_ctx, MARSHALL);
	prs_init(&rbuf, 0, mem_ctx, UNMARSHALL);

	init_paula_q_nt_logon_request(&q, cookie_in,
		domain, workstation, username,
		challenge, lm_response, nt_response_len, nt_response);

	if (!paula_io_q_nt_logon_request("", &q, &qbuf, 0))
		goto done;

	if (!paula_bind_and_request(mem_ctx, PAULA_NT_LOGON_REQUEST, &qbuf, &rbuf))
		goto done;

	if (!paula_io_r_nt_logon_request("", &r, &rbuf, 0))
		goto done;

	ret = True;

	*cookie_out = r.cookie;
	*status = r.status;
	*sub_status = r.sub_status;
	*result = r.result;

	memcpy(user_session_key, r.user_session_key, 16);
	memcpy(lm_session_key, r.lm_session_key, 8);

done:
	prs_mem_free(&qbuf);
	prs_mem_free(&rbuf);

	ZERO_STRUCT(q);
	ZERO_STRUCT(r);

	return;
}

static NTSTATUS check_paula_security(const struct auth_context *auth_context,
				       void *my_private_data, 
				       TALLOC_CTX *mem_ctx,
				       const auth_usersupplied_info *user_info, 
				       auth_serversupplied_info **server_info)
{
	uint32 cookie = 0;
	uint32 status, sub_status, result;
	uchar user_session_key[16];
	uchar lm_session_key[8];
	NTSTATUS nt_status;

	*server_info = NULL;

	if (user_info == NULL) {
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (auth_context == NULL) {
		DEBUG(3, ("check_paula_security: no authentication context provided for user %s\n",
			user_info->internal_username.str));
		return NT_STATUS_UNSUCCESSFUL;
	}

	if (auth_context->challenge.length != 8) {
		DEBUG(1, ("check_paula_security: cannot parse %d byte challenge\n",
			auth_context->challenge.length));
		return NT_STATUS_BUFFER_TOO_SMALL;
	}

	if (user_info->lm_resp.length != 0 && user_info->lm_resp.length != 24) {
		DEBUG(1, ("check_paula_security: cannot parse %d byte LM response\n",
			user_info->lm_resp.length));
		return NT_STATUS_BUFFER_TOO_SMALL;
	}

	paula_nt_logon_request(mem_ctx,
		cookie,
		user_info->domain.str,
		user_info->wksta_name.str,
		user_info->smb_name.str,
		auth_context->challenge.data,
		user_info->lm_resp.data,
		user_info->nt_resp.length,
		user_info->nt_resp.data,
		&cookie,
		&status,
		&sub_status,
		&result,
		user_session_key,
		lm_session_key);
	DEBUG(1, ("check_paula_security: paula_nt_logon_request returned result=%d status=%08x sub_status=%08x\n",
		(unsigned int)result, (unsigned int)status, (unsigned int)status));

	NT_STATUS_V(nt_status) = status;

	if (NT_STATUS_IS_OK(nt_status)) {
		NT_STATUS_V(nt_status) = sub_status;
	}

	if (NT_STATUS_IS_OK(nt_status)) {
		if (result == 0)
			nt_status = NT_STATUS_LOGON_FAILURE;
	}

	if (NT_STATUS_IS_OK(nt_status)) {
		struct passwd *pwd;

		pwd = Get_Pwnam(user_info->internal_username.str);
		if (pwd == NULL) {
			DEBUG(1, ("check_paula_security: could not find local account for %s\n",
				user_info->internal_username.str));
			nt_status = NT_STATUS_NO_SUCH_USER;
		} else {
			make_server_info_pw(server_info, pwd->pw_name, pwd);

			(*server_info)->user_session_key = data_blob(user_session_key, sizeof(user_session_key));
			(*server_info)->lm_session_key = data_blob(lm_session_key, sizeof(lm_session_key));
		}
	}

	memset(user_session_key, 0, sizeof(user_session_key));
	memset(lm_session_key, 0, sizeof(lm_session_key));

	return nt_status;
}

/* module initialisation */
static NTSTATUS auth_init_paula(struct auth_context *auth_context, const char *param, auth_methods **auth_method) 
{
	if (!make_auth_methods(auth_context, auth_method)) {
		return NT_STATUS_NO_MEMORY;
	}

	(*auth_method)->name = "paula";
	(*auth_method)->auth = check_paula_security;

	return NT_STATUS_OK;
}

NTSTATUS init_module(void)
{
	return smb_register_auth(AUTH_INTERFACE_VERSION, "paula", auth_init_paula);
}

