/*
 *  ncalrpc funnel: shunt data between SAMBA and FreeDCE
 *  Copyright (C) 2001-2004 PADL Software Pty Ltd. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 */

#ifdef NCALRPC_FUNNEL

#include <includes.h>

typedef struct ncalrpc_funnel {
	struct pipes_struct sp;
	int sd;
} ncalrpc_funnel;

typedef struct rpc_np_sec_context_t {
	uint32 Length;
	uint32 Version;
	uint32 UserNameLength;
	char *UserName;
	uint32 DomainNameLength;
	char *DomainName;
	uint32 WorkstationLength;
	char *Workstation;
	uint32 SessionKeyLength;
	char *SessionKey;
} rpc_np_sec_context_t, *rpc_np_sec_context_p_t;

/*
 * NCALRPC_PIPE_DIR should be a domain socket bound to
 * accept a ncalrpc connection with an additional round
 * trip at the front end, in which a simply serialized
 * rpc_np_sec_context_t is passed.
 */
#define NCALRPC_PIPE_DIR "/usr/xad/var/rpc/PIPE"

/**
 * Open up a connection to the ncalrpc server. We must
 * send a rpc_np_sec_context_t before we can treat it
 * as an ordinary ncalrpc server.
 */
static BOOL ncalrpc_open_funnel_socket(ncalrpc_funnel *p)
{
	struct sockaddr_un sunaddr;
	struct stat st;

	if (p == NULL) {
		return False;
	}

	ZERO_STRUCT(sunaddr);
	sunaddr.sun_family = AF_UNIX;

	if (strequal(p->sp.name, "lsarpc") ||
	    strequal(p->sp.name, "samr") ||
	    strequal(p->sp.name, "NETLOGON")) {
		snprintf(sunaddr.sun_path, sizeof(sunaddr.sun_path),
			"%s/lsass", NCALRPC_PIPE_DIR);
	} else {
		snprintf(sunaddr.sun_path, sizeof(sunaddr.sun_path),
			"%s/%s", NCALRPC_PIPE_DIR, p->sp.name);
	}

	become_root();
	if (lstat(sunaddr.sun_path, &st) == -1) {
		unbecome_root();
		DEBUG(0, ("ncalrpc_open_funnel_socket: could not access %s\n", sunaddr.sun_path));
		return False;
	}

	if (!S_ISSOCK(st.st_mode)) {
		unbecome_root();
		DEBUG(0, ("ncalrpc_open_funnel_socket: %s not socket\n", sunaddr.sun_path));
		return False;	
	}

	p->sd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (p->sd == -1) {
		unbecome_root();
		DEBUG(0, ("ncalrpc_open_funnel_socket: could not open socket - %s\n", strerror(errno)));
		return False;
	}

	if (connect(p->sd, (struct sockaddr *) &sunaddr, sizeof (sunaddr)) == -1) {
		unbecome_root();
		DEBUG(0, ("ncalrpc_open_funnel_socket: could not connect to socket - %s\n", strerror(errno)));
		return False;
	}

	unbecome_root();

	return True;
}

/**
 * Close a connection to the ncalrpc server.
 */
static void ncalrpc_close_funnel_socket(ncalrpc_funnel *p)
{
	if (p == NULL) {
		return;
	}

	if (p->sd == -1) {
		return;
	}

	close(p->sd);
	p->sd = -1;
}

static BOOL ncalrpc_send_sec_ctx(ncalrpc_funnel *p, connection_struct *conn, user_struct *vuser)
{
	rpc_np_sec_context_t ctx;
	ssize_t offset, bufsiz;
	char *buf;

	ZERO_STRUCT(ctx);

	ctx.Version = 2;

	ctx.WorkstationLength = sizeof("\\\\") - 1 + strlen(conn->client_address);
	ctx.Workstation = conn->client_address; /* add \\ later */

	if (vuser != NULL) {
		ctx.UserNameLength = strlen(vuser->user.smb_name);
		ctx.UserName = vuser->user.smb_name;
		ctx.DomainNameLength = strlen(vuser->user.domain);
		ctx.DomainName = vuser->user.domain;
		ctx.SessionKeyLength = vuser->session_key.length;
		ctx.SessionKey = vuser->session_key.data;
	}

	ctx.Length = sizeof(ctx.Version) +
		sizeof(ctx.UserNameLength) + ctx.UserNameLength +
		sizeof(ctx.DomainNameLength) + ctx.DomainNameLength +
		sizeof(ctx.WorkstationLength) + ctx.WorkstationLength +
		sizeof(ctx.SessionKeyLength) + ctx.SessionKeyLength;
	bufsiz = sizeof(ctx.Length) + ctx.Length;
	buf = malloc(bufsiz);
	if (buf == NULL) {
		DEBUG(0, ("ncalrpc_send_sec_ctx: could not allocate buffer\n"));
		return False;
	}

	offset = 0;
	SIVAL(buf, offset, ctx.Length);
	offset += sizeof(ctx.Length);

	SIVAL(buf, offset, ctx.Version);
	offset += sizeof(ctx.Version);

	SIVAL(buf, offset, ctx.UserNameLength);
	offset += sizeof(ctx.UserNameLength);
	if (ctx.UserName != NULL) {
		memcpy(buf + offset, ctx.UserName, ctx.UserNameLength);
	}
	offset += ctx.UserNameLength;

	SIVAL(buf, offset, ctx.DomainNameLength);
	offset += sizeof(ctx.DomainNameLength);
	if (ctx.DomainName != NULL) {
		memcpy(buf + offset, ctx.DomainName, ctx.DomainNameLength);
	}
	offset += ctx.DomainNameLength;

	/* Prepend \\ to WorkstationLength; let's make that convention. */
	SIVAL(buf, offset, ctx.WorkstationLength);
	offset += sizeof(ctx.WorkstationLength);
	memset(buf + offset, '\\', 2);
	offset += 2;
	memcpy(buf + offset, ctx.Workstation, ctx.WorkstationLength - 2);
	offset += ctx.WorkstationLength - 2;

	SIVAL(buf, offset, ctx.SessionKeyLength);
	offset += sizeof(ctx.SessionKeyLength);
	if (ctx.SessionKey != NULL) {
		memcpy(buf + offset, ctx.SessionKey, ctx.SessionKeyLength);
	}
	offset += ctx.SessionKeyLength;

	dump_data(0, buf, bufsiz);

	if (ncalrpc_write_to_funnel(p, buf, bufsiz) != bufsiz) {
		DEBUG(0, ("ncalrpc_send_sec_ctx: could not send security context\n"));
		free(buf);
		return False;
	}
	free(buf);

	return True;
}

void *ncalrpc_make_funnel(char *pipe_name, connection_struct *conn, uint16 vuid)
{
	user_struct *vuser;
	ncalrpc_funnel *p;

	DEBUG(4, ("ncalrpc_make_funnel: pipe %s vuid %d\n", pipe_name, vuid));

	vuser = get_valid_user_struct(vuid);

	p = (ncalrpc_funnel *)malloc(sizeof(*p));
	if (p == NULL) {
		DEBUG(0, ("ncalrpc_make_funnel: out of memory\n"));
		return NULL;
	}

	ZERO_STRUCTP(p);
	p->sd = -1;

	p->sp.mem_ctx = talloc_init("ncalrpc_funnel mem_ctx");
	if (p->sp.mem_ctx == NULL) {
		DEBUG(0, ("ncalrpc_make_funnel: talloc_init failed\n"));
		SAFE_FREE(p);
		return NULL;
	}

	if (!init_pipe_handle_list(&p->sp, pipe_name)) {
		DEBUG(0, ("ncalrpc_make_funnel: init_pipe_handle_list failed\n"));
		talloc_destroy(p->sp.mem_ctx);
		SAFE_FREE(p);
		return NULL;
	}

	if (!prs_init(&p->sp.in_data.data, MAX_PDU_FRAG_LEN, p->sp.mem_ctx, MARSHALL)) {
		DEBUG(0, ("ncalrpc_make_funnel: prs_init failed\n"));
		talloc_destroy(p->sp.mem_ctx);
		SAFE_FREE(p);
		return NULL;
	}

	fstrcpy(p->sp.name, pipe_name);

	if (ncalrpc_open_funnel_socket(p) == False) {
		DEBUG(0, ("ncalrpc_make_funnel: could not open socket for funnel pipe %s\n", pipe_name));
		SAFE_FREE(p);
		return NULL;
	}

	p->sp.conn = conn;
	p->sp.conn->num_files_open++;
	p->sp.vuid = vuid;

	p->sp.ntlmssp_chal_flags = 0;
	p->sp.ntlmssp_auth_validated = False;
	p->sp.ntlmssp_auth_requested = False;

	p->sp.pipe_bound = False;
	p->sp.fault_state = False;
	p->sp.endian = RPC_LITTLE_ENDIAN;

	ZERO_STRUCT(p->sp.pipe_user);

	p->sp.pipe_user.uid = -1;
	p->sp.pipe_user.gid = -1;

	if (ncalrpc_send_sec_ctx(p, conn, vuser) == False) {
		ncalrpc_close_funnel(p);
		return NULL;
	}

	p->sp.in_data.pdu_needed_len = 0;
	p->sp.in_data.pdu_received_len = 0;

	p->sp.out_data.current_pdu_len = 0;
	p->sp.out_data.current_pdu_sent = 0;
	p->sp.out_data.data_sent_length = 0;

	prs_init(&p->sp.out_data.rdata, 0, p->sp.mem_ctx, MARSHALL);

	DEBUG(4, ("Created ncalrpc funnel pipe %s\n", pipe_name));

	return p;
}

/**
 * Read shim. 
 */
ssize_t ncalrpc_read_from_funnel(void *np_conn,
	char *data,
	size_t n,
	BOOL *is_data_outstanding)
{
	ncalrpc_funnel *p = (ncalrpc_funnel *)np_conn;
	ssize_t ret, selrtn;
	fd_set fds;
	struct timeval timeout;

	*is_data_outstanding = False;

	if (p->sd < 0) {
	   return -1;
	}

	/* Check whether n > MAX_PDU_FRAG_LEN??? */
	ret = sys_read(p->sd, data, n);

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	FD_ZERO(&fds);
	FD_SET(p->sd, &fds);

	selrtn = sys_select(p->sd + 1, &fds, NULL, NULL, &timeout);
	if (selrtn > 0 && FD_ISSET(p->sd, &fds)) {
		*is_data_outstanding = True;
	}

	DEBUG(3, ("ncalrpc_funnel: read %d bytes from funnel (maximum %d); %smore data to read\n", ret, n, *is_data_outstanding ? "" : "no "));

	return ret;
}

/**
 * Write shim. 
 */
ssize_t ncalrpc_write_to_funnel(void *np_conn,
	char *data,
	size_t n)
{
	ncalrpc_funnel *p = (ncalrpc_funnel *)np_conn;
	ssize_t ret;

	if (p->sd < 0) {
		return -1;
	}

	ret = write_socket(p->sd, data, n);

	DEBUG(3, ("ncalrpc_write_to_funnel: wrote %d of %d bytes to funnel\n", ret, n));

	return ret;
}

/**
 * Close the RPC pipe handle.
 */
BOOL ncalrpc_close_funnel(void *np_conn)
{
	ncalrpc_funnel *p = (ncalrpc_funnel *)np_conn;
	if (p == NULL) {
		DEBUG(0, ("ncalrpc_close_funnel: invalid pipe\n"));
		return False;
	}
	prs_mem_free(&p->sp.out_data.rdata);
	prs_mem_free(&p->sp.in_data.data);

	if (p->sp.mem_ctx != NULL) {
		talloc_destroy(p->sp.mem_ctx);
	}

	delete_nt_token(&p->sp.pipe_user.nt_user_token);

	SAFE_FREE(p->sp.pipe_user.groups);
	p->sp.conn->num_files_open--;

	ncalrpc_close_funnel_socket(p);

	ZERO_STRUCT(p);
	SAFE_FREE(p);

	DEBUG(3, ("ncalrpc_close_funnel: closed funnel\n"));

	return True;
}

#endif /* NCALRPC_FUNNEL */

