/* 
 * This module implements kerberos5 authentication. Both native kerberos 
 * (via GSS API) and password-based methods are supported. Hacked browser is 
 * needed if you want to use native kerberos5 protocol (a patch for the Mozilla
 * browser can be found at http://meta.cesnet.cz/software/heimdal/)
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"

#include <gssapi.h>

#include <krb5.h>
#include <kafs.h>

module MODULE_VAR_EXPORT gss_krb5_auth_module;

extern krb5_keytab gssapi_krb5_keytab;

typedef struct {
  /* Directives for the GSSAPI part */
  char  *gss_krb5_keytab;  
  int    gss_save_creds;
  char  *gss_krb5_cells;
  int    use_gss_auth;
  char  *gss_krb5_realms;

  /* Directives for the password verification part */
  char *krb5_auth_realm;
  char *krb5_keytab;
  int krb5_save_creds;
  int krb5_forwardable;
  char *krb5_cells;
} gss_krb5_config_rec;

static const char *
krb5_save_cells(cmd_parms *, gss_krb5_config_rec *, char *);

static const char *
gss_save_cells(cmd_parms *, gss_krb5_config_rec *, char *);

static const char *
gss_save_realms(cmd_parms *, gss_krb5_config_rec *, char *);

static const char *
krb5_save_realms(cmd_parms *, gss_krb5_config_rec *, char *);

static const command_rec gss_krb5_cmds[] =
{
  /* Directives for the GSSAPI part */ 
  { "GssKrb5Keytab", ap_set_file_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, gss_krb5_keytab),
    OR_AUTHCFG, TAKE1, "Kerberos v5 keytab." },
  { "GssKrb5SaveCredentials", ap_set_flag_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, gss_save_creds),
    OR_AUTHCFG, FLAG, "Save the v5 credential cache?" },
  { "GssKrb5AFSCells", gss_save_cells, NULL,
     OR_AUTHCFG, RAW_ARGS, "Create pag and AFS tokens for cells" },
  { "GssSaveCredentials", ap_set_flag_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, gss_save_creds),
    OR_AUTHCFG, FLAG, "Save the v5 credential cache?" },
  { "GssAuth", ap_set_flag_slot,
     (void*)XtOffsetOf(gss_krb5_config_rec, use_gss_auth),
     OR_AUTHCFG, FLAG, "Whether to use GSS-API authentication" },
  { "GssKrb5AuthRealms", gss_save_realms, NULL,
     OR_AUTHCFG, RAW_ARGS, "Kerberos realm(s) in which to authenticate users" },

  /* Directives for the password verification part */
  { "KrbAuthRealm", krb5_save_realms, NULL,
    OR_AUTHCFG, RAW_ARGS, "Kerberos realm in which to authenticate users." },

   { "Krb5Keytab", ap_set_file_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, krb5_keytab),
    OR_AUTHCFG, TAKE1, "Kerberos v5 keytab." },

   { "Krb5SaveCredentials", ap_set_flag_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, krb5_save_creds),
    OR_AUTHCFG, FLAG, "Save the v5 credential cache?" },

   { "Krb5Forwardable", ap_set_flag_slot,
    (void*)XtOffsetOf(gss_krb5_config_rec, krb5_forwardable),
    OR_AUTHCFG, FLAG, "Set tickets to be forwardable?" },

   { "Krb5AFSCells", krb5_save_cells, NULL,
    OR_AUTHCFG, RAW_ARGS, "Create pag and AFS tokens for cells" },   

  { NULL, NULL, NULL, 0, 0, NULL}
};

static const char *
gss_save_cells(cmd_parms *cmd, gss_krb5_config_rec *sec, char *arg)
{
   sec->gss_krb5_cells = ap_pstrdup(cmd->pool,arg);
   return NULL;
}

static const char *
krb5_save_cells(cmd_parms *cmd, gss_krb5_config_rec *sec, char *arg)
{
   sec->krb5_cells = ap_pstrdup(cmd->pool,arg);
   return NULL;
}

static const char *
gss_save_realms(cmd_parms *cmd, gss_krb5_config_rec *sec, char *arg)
{
   sec->gss_krb5_realms = ap_pstrdup(cmd->pool,arg);
   return NULL;
}

static const char *
krb5_save_realms(cmd_parms *cmd, gss_krb5_config_rec *sec, char *arg)
{
   sec->krb5_auth_realm = ap_pstrdup(cmd->pool,arg);
   return NULL;
}

static void
krb5_cache_cleanup(void *data)
{
   krb5_context context;
   krb5_ccache  cache;
   krb5_error_code problem;
   char *cache_name = (char *) data;

   problem = krb5_init_context(&context);
   if (problem) {
      ap_log_error(APLOG_MARK, APLOG_ERR, NULL, "krb5_init_context() failed");
      return;
   }

   problem = krb5_cc_resolve(context, cache_name, &cache);
   if (problem) {
      ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
                   "krb5_cc_resolve() failed (%s: %s)",
	           cache_name, krb5_get_err_text(context, problem)); 
      return;
   }

   krb5_cc_destroy(context, cache);
   krb5_free_context(context);
}

static void krb5_dummy_cleanup(void *data) {}

static void krb5_unlog_cleanup(void *data)
{
   if (k_hasafs())
      k_unlog();
}

static void
note_auth_failure(request_rec *r, const gss_krb5_config_rec *conf, int spnego)
{
   const char *auth_type = NULL;
   const char *auth_name = NULL;

   /* get the type specified in .htaccess */
   auth_type = ap_auth_type(r);

   /* get the user realm specified in .htaccess */
   auth_name = ap_auth_name(r);

   /* XXX should the WWW-Authenticate header be cleared first? */
   if (conf->use_gss_auth)
      ap_table_add(r->err_headers_out, "WWW-Authenticate", spnego ? "Negotiate " : "GSS-Negotiate ");
   if (auth_type && strncasecmp(auth_type, "KerberosV5", 10) == 0)
      ap_table_add(r->err_headers_out, "WWW-Authenticate",
	           ap_pstrcat(r->pool, "Basic realm=\"", auth_name, "\"", NULL));
}

static const char *
get_gss_error(pool *p, OM_uint32 error_status, char *prefix)
{
   OM_uint32 maj_stat, min_stat;
   OM_uint32 msg_ctx = 0;
   gss_buffer_desc status_string;
   char buf[1024];
   size_t len;

   snprintf(buf, sizeof(buf), "%s: ", prefix);
   len = strlen(buf);
   do {
      maj_stat = gss_display_status (&min_stat,
	                             error_status,
				     GSS_C_MECH_CODE,
				     GSS_C_NO_OID,
				     &msg_ctx,
				     &status_string);
      if (sizeof(buf) > len + status_string.length + 1) {
         sprintf(buf, "%s:", (char*) status_string.value);
         len += status_string.length;
      }
      gss_release_buffer(&min_stat, &status_string);
   } while (!GSS_ERROR(maj_stat) && msg_ctx != 0);

   return (ap_pstrdup(p, buf));
}

static void *
gss_krb5_dir_create_config(pool *p, char *dir)
{
  gss_krb5_config_rec *rec;

  rec = (gss_krb5_config_rec *) ap_pcalloc(p, sizeof(gss_krb5_config_rec));
  return rec;
}

static void
do_afs_log(krb5_context krb_ctx,
           request_rec *r,
	   krb5_ccache ccache,
	   char *afs_cells)
{
   if (afs_cells && k_hasafs()) {
      char *cells = ap_pstrdup(r->pool, afs_cells), *next;

      k_setpag();
      for (next=strtok(cells," \t"); next; next = strtok(NULL," \t"))
	  krb5_afslog(krb_ctx, ccache, next, NULL);
      ap_register_cleanup(r->pool, NULL, 
	                  krb5_unlog_cleanup, krb5_dummy_cleanup);
   }
}
static int
store_krb5_creds(krb5_context krb_ctx,
                 request_rec *r,
                 gss_krb5_config_rec *conf,
                 krb5_ccache delegated_cred)
{
   krb5_error_code problem;
   char *cache_name = NULL;
   krb5_ccache ccache = NULL;
   krb5_principal principal = NULL;

   if (delegated_cred == NULL)
      return OK; /* XXX */

   problem = krb5_cc_gen_new(krb_ctx, &krb5_fcc_ops, &ccache);
   if (problem) {
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	            "krb5_cc_gen_new() failed (%s)",
		    krb5_get_err_text(krb_ctx, problem));
      goto end;
   }

   problem = krb5_cc_get_principal(krb_ctx, delegated_cred, &principal);
   if (problem) {
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	            "krb5_cc_get_principal() failed (%s)",
		    krb5_get_err_text(krb_ctx, problem));
      goto end;
   }

   problem = krb5_cc_initialize(krb_ctx, ccache, principal);
   if (problem) {
       ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	             "krb5_cc_initialize() failed (%s)",
		     krb5_get_err_text(krb_ctx, problem));
       goto end;
   }

   problem = krb5_cc_copy_cache(krb_ctx, delegated_cred, ccache);
   if (problem) {
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	            "krb5_cc_copy_cache() failed (%s)",
		    krb5_get_err_text(krb_ctx, problem));
      goto end;
   }

   cache_name = ap_pstrdup(r->pool, krb5_cc_get_name(krb_ctx, ccache));
   ap_table_setn(r->subprocess_env, "KRB5CCNAME", cache_name);
   ap_register_cleanup(r->pool, cache_name, 
	               krb5_cache_cleanup, krb5_dummy_cleanup);
     
   krb5_cc_close(krb_ctx, ccache);

end:
   if (principal)
      krb5_free_principal(krb_ctx, principal);
   if (ccache && problem)
      krb5_cc_destroy(krb_ctx, ccache);

   return (problem ? SERVER_ERROR : OK);
}

static int 
authenticate_user_gss(request_rec *r, 
                      gss_krb5_config_rec *conf,
                      const char *auth_line,
		      int spnego)
{
  OM_uint32 major_status, minor_status, minor_status2;
  gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;
  gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
  const char *auth_param = NULL;
  krb5_context krb_ctx = NULL;
  int ret;
  gss_name_t client_name = GSS_C_NO_NAME;
  gss_ctx_id_t gss_context = GSS_C_NO_CONTEXT;
  gss_cred_id_t delegated_cred = GSS_C_NO_CREDENTIAL;
  char *p;

  krb5_init_context(&krb_ctx);
  if (conf->gss_krb5_keytab &&
      krb5_kt_resolve(krb_ctx, conf->gss_krb5_keytab, &gssapi_krb5_keytab)) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	             "krb5_kt_resolve() failed");
	ret = SERVER_ERROR;
	goto end;
  }

  /* ap_getword() shifts parameter */
  auth_param = ap_getword_white(r->pool, &auth_line);
  if (auth_param == NULL) {
     ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	           "No Authorization parameter from client");
     ret = HTTP_UNAUTHORIZED;
     goto end;
  }

  /* XXX use ap_uudecode() */
  input_token.length = ap_base64decode_len(auth_param);
  input_token.value = malloc(input_token.length);
  if (input_token.value == NULL) {
     ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	   	   "Not enough memory");
     ret = SERVER_ERROR;
     goto end;
  }
  input_token.length = ap_base64decode(input_token.value, auth_param);

  major_status = gss_accept_sec_context(&minor_status,
	                                &gss_context,
					GSS_C_NO_CREDENTIAL,
					&input_token,
					GSS_C_NO_CHANNEL_BINDINGS,
					&client_name,
					NULL,
					&output_token,
					NULL,
					NULL,
					&delegated_cred);
  if (output_token.length) {
     char *token = NULL;
     
     /* XXX use ap_uuencode() */
     token = malloc(ap_base64encode_len(output_token.length));
     if (token == NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	             "Not enough memory");
        ret = SERVER_ERROR;
	gss_release_buffer(&minor_status2, &output_token);
	goto end;
     }
     ap_base64encode(token, output_token.value, output_token.length);
     ap_table_set(r->err_headers_out, "WWW-Authenticate",
	          ap_pstrcat(r->pool, spnego ? "Negotiate " : "GSS-Negotiate ", token, NULL));
     free(token);
     gss_release_buffer(&minor_status2, &output_token);
  }

  if (GSS_ERROR(major_status)) {
     ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	           "%s", get_gss_error(r->pool, minor_status,
		                       "gss_accept_sec_context() failed"));
     ret = HTTP_UNAUTHORIZED;
     goto end;
  }

  if (major_status & GSS_S_CONTINUE_NEEDED) {
     ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	           "only one authentication iteration allowed"); 
     ret = HTTP_UNAUTHORIZED;
     goto end;
  }

  major_status = gss_export_name(&minor_status, client_name, &output_token);
  gss_release_name(&minor_status, &client_name); 
  if (GSS_ERROR(major_status)) {
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, r,
	          "%s", get_gss_error(r->pool, minor_status, 
		                      "gss_export_name() failed"));
    ret = SERVER_ERROR;
    goto end;
  }

  /* If the user comes from a realm specified by configuration don't include
      its realm name in the username so that the authorization routine could
      work for both Password-based and Ticket-based authentication. It's
      administrators responsibility to include only such realm that have
      unified principal instances, i.e. if the same principal name occures in
      multiple realms, it must be always assigned to a single user.
  */    
  r->connection->ap_auth_type = "Negotiate";
  r->connection->user = ap_pstrdup(r->pool, output_token.value);
  p = strchr(r->connection->user, '@');
  if (p != NULL) {
     const char *realms = conf->gss_krb5_realms;

     while (realms && *realms) {
	if (strcmp(p+1, ap_getword_white(r->pool, &realms)) == 0) {
	   *p = '\0';
	   break;
	}
     }
  }

  gss_release_buffer(&minor_status, &output_token);

  /* This should be only done if afs token are requested or gss_save creds is 
   * specified */
  /* gss_export_cred() from the GGF GSS Extensions could be used */
  if (delegated_cred != GSS_C_NO_CREDENTIAL &&
      (conf->gss_save_creds || (conf->gss_krb5_cells && k_hasafs()))) {	
     struct krb5_ccache_data ccache;

     major_status = gss_krb5_copy_ccache(&minor_status, delegated_cred, &ccache);
     if (GSS_ERROR(major_status)) {
	ret = SERVER_ERROR;
	goto end;
     }

     do_afs_log(krb_ctx, r, &ccache, conf->gss_krb5_cells);

     ret = store_krb5_creds(krb_ctx, r, conf, &ccache);
     if (ret) {
	krb5_cc_close(krb_ctx, &ccache);
	goto end;
     }

     krb5_cc_close(krb_ctx, &ccache);
  }
  ret = OK;

end:
  if (delegated_cred)
     gss_release_cred(&minor_status, &delegated_cred);

  if (output_token.length) 
     gss_release_buffer(&minor_status, &output_token);

  if (client_name != GSS_C_NO_NAME)
     gss_release_name(&minor_status, &client_name);

  if (gss_context != GSS_C_NO_CONTEXT)
     gss_delete_sec_context(&minor_status, &gss_context, GSS_C_NO_BUFFER);

  krb5_free_context(krb_ctx);

  if (ret == HTTP_UNAUTHORIZED)
     note_auth_failure(r, conf, spnego);

  return ret;
}

static int
authenticate_user_pwd(request_rec *r,
                      gss_krb5_config_rec *conf,
		      const char *auth_line)
{
   const char      *sent_pw = NULL;
   const char      *realms = NULL;
   krb5_context    kcontext;
   krb5_error_code code;
   krb5_principal  princ = NULL;
   char            errstr[512];
   krb5_ccache     ccache = NULL;
   int             ret;

   krb5_init_context(&kcontext);
   
   sent_pw = ap_uudecode(r->pool, auth_line);
   r->connection->user = ap_getword (r->pool, &sent_pw, ':');
   r->connection->ap_auth_type = "Basic";
   
   /* do not allow user to override realm setting of server */
   if (strchr(r->connection->user,'@')) {
      ap_log_reason ("specifying realm in user name is prohibited", r->uri, r);
      return  HTTP_UNAUTHORIZED;
   }

   code = krb5_cc_gen_new(kcontext, &krb5_mcc_ops, &ccache);
   if (code) {
      snprintf(errstr, sizeof(errstr), "krb5_cc_gen_new(): %.100s",
	       krb5_get_err_text(kcontext, code));
      ap_log_reason (errstr, r->uri, r);
      ret = SERVER_ERROR;
      goto end;
   }

   if (conf->krb5_keytab)
      kcontext->default_keytab = conf->krb5_keytab;

   realms = conf->krb5_auth_realm;

   do {
      code = 0;
      if (realms)
	 code = krb5_set_default_realm(kcontext,
	                               ap_getword_white(r->pool, &realms));
      if (code)
	 continue;

      code = krb5_parse_name(kcontext, r->connection->user, &princ);
      if (code)
	 continue;

      code = krb5_verify_user(kcontext, princ, ccache, sent_pw, 1, "khttp");
      if (code == 0)
	 break;
   } while (realms && *realms);

   memset((char *)sent_pw, 0x00, strlen(sent_pw));

   if (code) {
      snprintf(errstr, sizeof(errstr), "Verifying krb5 password failed: %s",
               krb5_get_err_text(kcontext, code));
      ap_log_reason (errstr, r->uri, r);
      ret = HTTP_UNAUTHORIZED;
      goto end;
   }

   if (conf->krb5_save_creds) {
      do_afs_log(kcontext, r, ccache, conf->krb5_cells);
      ret = store_krb5_creds(kcontext, r, conf, ccache);
      if (ret)
         goto end;
   }

   ret = OK;

end:
   if (princ)
      krb5_free_principal(kcontext, princ);
   if (ccache)
      krb5_cc_destroy(kcontext, ccache);
   krb5_free_context(kcontext);

   return ret;
}

static int
gss_krb5_authenticate_user(request_rec *r)
{
  gss_krb5_config_rec *conf =
     (gss_krb5_config_rec *) ap_get_module_config(r->per_dir_config,
						  &gss_krb5_auth_module);
  const char *auth_type = NULL;
  const char *auth_line = NULL;
  const char *type = NULL;
  int ret;

  /* get the type specified in .htaccess */
  type = ap_auth_type(r);

  /* Apache doesn't allow multiple values in AuthType within one directory.
   * But since we want to support both the method we added a new configuration
   * option (UseGSS) to turn on GSS authentication. */
  if (!conf->use_gss_auth &&
      (type == NULL || strncasecmp(type, "KerberosV5", 10) != 0)) {
     return DECLINED;
  }

  /* get what the user sent us in the HTTP header */
  auth_line = ap_table_get (r->headers_in, "Authorization");
  if (!auth_line) {
      note_auth_failure(r, conf, TRUE);
      return HTTP_UNAUTHORIZED;
  }
  auth_type = ap_getword_white(r->pool, &auth_line);

  if (conf->use_gss_auth) {
     if (strncasecmp(auth_type, "GSS-Negotiate", 13) == 0) {
	ret = authenticate_user_gss(r, conf, auth_line, FALSE);
     } else if (strncasecmp(auth_type, "Negotiate", 9) == 0) {
	ret = authenticate_user_gss(r, conf, auth_line, TRUE);
     }
  } else
     if (type != NULL && strncasecmp(type, "KerberosV5", 10) == 0 &&
	 strncasecmp(auth_type, "Basic", 5) == 0) {
	ret = authenticate_user_pwd(r, conf, auth_line);
     }
     else {
	ret = HTTP_UNAUTHORIZED;
     }

  if (ret == HTTP_UNAUTHORIZED)
     note_auth_failure(r, conf, TRUE);

  return ret;
}


static int 
gss_krb5_check_user(request_rec *r)
{
  gss_krb5_config_rec *conf =
     (gss_krb5_config_rec *) ap_get_module_config(r->per_dir_config,
						&gss_krb5_auth_module);
  char *user = r->connection->user;
  int m = r->method_number;
  int method_restricted = 0;
  register int x;
  const char *t, *w;
  const require_line *require;
  const array_header *required_users;

  required_users = ap_requires(r);
  if (required_users == NULL)
    return OK;

  require = (require_line *) required_users->elts;

  for (x = 0; x < required_users->nelts; x++) {
     if (! (require[x].method_mask & (1 << m)))
	continue;

     method_restricted = 1;

     t = require[x].requirement;
     w = ap_getword_white(r->pool, &t);
     if (strcmp(w, "valid-user") == 0) {
	return OK;
     }

     if (strcmp(w, "user") == 0) {
	while (t[0] != '\0') {
	   w = ap_getword_conf(r->pool, &t);
           if (strcmp(user, w) == 0) 
              return OK;
	}
     }
  }

  if (! method_restricted)
     return OK;

  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
	"GSS & Krb5 module: access to %s failed, reason: user %s not allowed access",
	r->uri, user);

  note_auth_failure(r, conf, TRUE);

  return AUTH_REQUIRED;
}

module MODULE_VAR_EXPORT gss_krb5_auth_module =
{
  STANDARD_MODULE_STUFF, 
  NULL,	                      /* initializer */
  gss_krb5_dir_create_config, /* dir config creater */
  NULL,                       /* dir merger --- default is to override */
  NULL,                       /* server config */
  NULL,                       /* merge server config */
  gss_krb5_cmds,              /* command table */
  NULL,                       /* handlers */
  NULL,                       /* filename translation */
  gss_krb5_authenticate_user, /* check_user_id */
  gss_krb5_check_user,        /* check auth */
  NULL,                       /* check access */
  NULL,                       /* type_checker */
  NULL,                       /* fixups */
  NULL,                       /* logger */
  NULL,                       /* header parser */
  NULL,                       /* child_init */
  NULL,                       /* child_exit */
  NULL                        /* post read-request */
};
